@@ -35,6 +35,7 @@ array set ::redis::addr {}
 array set ::redis::blocking {}
 array set ::redis::deferred {}
 array set ::redis::readraw {}
+array set ::redis::attributes {} ;# Holds the RESP3 attributes from the last call
 array set ::redis::reconnect {}
 array set ::redis::tls {}
 array set ::redis::callback {}
@@ -66,33 +67,6 @@ proc redis {{server 127.0.0.1} {port 6379} {defer 0} {tls 0} {tlsoptions {}} {re
     interp alias {} ::redis::redisHandle$id {} ::redis::__dispatch__ $id
 }
 
-# On recent versions of tcl-tls/OpenSSL, reading from a dropped connection
-# results with an error we need to catch and mimic the old behavior.
-proc ::redis::redis_safe_read {fd len} {
-    if {$len == -1} {
-        set err [catch {set val [read $fd]} msg]
-    } else {
-        set err [catch {set val [read $fd $len]} msg]
-    }
-    if {!$err} {
-        return $val
-    }
-    if {[string match "*connection abort*" $msg]} {
-        return {}
-    }
-    error $msg
-}
-
-proc ::redis::redis_safe_gets {fd} {
-    if {[catch {set val [gets $fd]} msg]} {
-        if {[string match "*connection abort*" $msg]} {
-            return {}
-        }
-        error $msg
-    }
-    return $val
-}
-
 # This is a wrapper to the actual dispatching procedure that handles
 # reconnection if needed.
 proc ::redis::__dispatch__ {id method args} {
@@ -111,7 +85,7 @@ proc ::redis::__dispatch__raw__ {id method argv} {
     set fd $::redis::fd($id)
 
     # Reconnect the link if needed.
-    if {$fd eq {}} {
+    if {$fd eq {} && $method ne {close}} {
         lassign $::redis::addr($id) host port
         if {$::redis::tls($id)} {
             set ::redis::fd($id) [::tls::socket $host $port]
@@ -132,6 +106,7 @@ proc ::redis::__dispatch__raw__ {id method argv} {
         set argv [lrange $argv 0 end-1]
     }
     if {[info command ::redis::__method__$method] eq {}} {
+        catch {unset ::redis::attributes($id)}
         set cmd "*[expr {[llength $argv]+1}]\r\n"
         append cmd "$[string length $method]\r\n$method\r\n"
         foreach a $argv {
@@ -173,6 +148,10 @@ proc ::redis::__method__read {id fd} {
     ::redis::redis_read_reply $id $fd
 }
 
+proc ::redis::__method__rawread {id fd len} {
+    return [read $fd $len]
+}
+
 proc ::redis::__method__write {id fd buf} {
     ::redis::redis_write $fd $buf
 }
@@ -188,6 +167,7 @@ proc ::redis::__method__close {id fd} {
     catch {unset ::redis::blocking($id)}
     catch {unset ::redis::deferred($id)}
     catch {unset ::redis::readraw($id)}
+    catch {unset ::redis::attributes($id)}
     catch {unset ::redis::reconnect($id)}
     catch {unset ::redis::tls($id)}
     catch {unset ::redis::state($id)}
@@ -208,6 +188,14 @@ proc ::redis::__method__readraw {id fd val} {
     set ::redis::readraw($id) $val
 }
 
+proc ::redis::__method__readingraw {id fd} {
+    return $::redis::readraw($id)
+}
+
+proc ::redis::__method__attributes {id fd} {
+    set _ $::redis::attributes($id)
+}
+
 proc ::redis::redis_write {fd buf} {
     puts -nonewline $fd $buf
 }
@@ -219,8 +207,8 @@ proc ::redis::redis_writenl {fd buf} {
 }
 
 proc ::redis::redis_readnl {fd len} {
-    set buf [redis_safe_read $fd $len]
-    redis_safe_read $fd 2 ; # discard CR LF
+    set buf [read $fd $len]
+    read $fd 2 ; # discard CR LF
     return $buf
 }
 
@@ -266,11 +254,11 @@ proc ::redis::redis_read_map {id fd} {
 }
 
 proc ::redis::redis_read_line fd {
-    string trim [redis_safe_gets $fd]
+    string trim [gets $fd]
 }
 
 proc ::redis::redis_read_null fd {
-    redis_safe_gets $fd
+    gets $fd
     return {}
 }
 
@@ -293,7 +281,7 @@ proc ::redis::redis_read_reply {id fd} {
     }
 
     while {1} {
-        set type [redis_safe_read $fd 1]
+        set type [read $fd 1]
         switch -exact -- $type {
             _ {return [redis_read_null $fd]}
             : -
@@ -309,8 +297,8 @@ proc ::redis::redis_read_reply {id fd} {
             * {return [redis_multi_bulk_read $id $fd]}
             % {return [redis_read_map $id $fd]}
             | {
-                # ignore attributes for now (nowhere to store them)
-                redis_read_map $id $fd
+                set attrib [redis_read_map $id $fd]
+                set ::redis::attributes($id) $attrib
                 continue
             }
             default {
