@@ -24,10 +24,10 @@ proc assert_no_match {pattern value} {
     }
 }
 
-proc assert_match {pattern value} {
+proc assert_match {pattern value {detail ""}} {
     if {![string match $pattern $value]} {
         set context "(context: [info frame -1])"
-        error "assertion:Expected '$value' to match '$pattern' $context"
+        error "assertion:Expected '$value' to match '$pattern' $context $detail"
     }
 }
 
@@ -40,6 +40,12 @@ proc assert_failed {expected_err detail} {
      error "assertion:$expected_err $detail"
 }
 
+proc assert_not_equal {value expected {detail ""}} {
+    if {!($expected ne $value)} {
+        assert_failed "Expected '$value' not equal to '$expected'" $detail
+    }
+}
+
 proc assert_equal {value expected {detail ""}} {
     if {$expected ne $value} {
         assert_failed "Expected '$value' to be equal to '$expected'" $detail
@@ -76,23 +82,31 @@ proc assert_range {value min max {detail ""}} {
     }
 }
 
-proc assert_error {pattern code} {
+proc assert_error {pattern code {detail ""}} {
     if {[catch {uplevel 1 $code} error]} {
-        assert_match $pattern $error
+        assert_match $pattern $error $detail
     } else {
-        error "assertion:Expected an error but nothing was caught"
+        assert_failed "Expected an error matching '$pattern' but got '$error'" $detail
     }
 }
 
 proc assert_encoding {enc key} {
-    set dbg [r debug object $key]
-    assert_match "* encoding:$enc *" $dbg
+    if {$::ignoreencoding} {
+        return
+    }
+    set val [r object encoding $key]
+    assert_match $enc $val
 }
 
 proc assert_type {type key} {
     assert_equal $type [r type $key]
 }
 
+proc assert_refcount {ref key} {
+    set val [r object refcount $key]
+    assert_equal $ref $val
+}
+
 # Wait for the specified condition to be true, with the specified number of
 # max retries and delay between retries. Otherwise the 'elsescript' is
 # executed.
@@ -112,46 +126,57 @@ proc wait_for_condition {maxtries delay e _else_ elsescript} {
     }
 }
 
-proc test {name code {okpattern undefined} {options undefined}} {
+# try to match a value to a list of patterns that is either regex, or plain sub-string
+proc search_pattern_list {value pattern_list {substr false}} {
+    set n 0
+    foreach el $pattern_list {
+        if {[string length $el] > 0 && ((!$substr && [regexp -- $el $value]) || ($substr && [string match $el $value]))} {
+            return $n
+        }
+        incr n
+    }
+    return -1
+}
+
+proc test {name code {okpattern undefined} {tags {}}} {
     # abort if test name in skiptests
-    if {[lsearch $::skiptests $name] >= 0} {
+    if {[search_pattern_list $name $::skiptests] >= 0} {
         incr ::num_skipped
         send_data_packet $::test_server_fd skip $name
         return
     }
 
     # abort if only_tests was set but test name is not included
-    if {[llength $::only_tests] > 0 && [lsearch $::only_tests $name] < 0} {
+    if {[llength $::only_tests] > 0 && [search_pattern_list $name $::only_tests] < 0} {
         incr ::num_skipped
         send_data_packet $::test_server_fd skip $name
         return
     }
 
-    # check if tagged with at least 1 tag to allow when there *is* a list
-    # of tags to allow, because default policy is to run everything
-    if {[llength $::allowtags] > 0} {
-        set matched 0
-        foreach tag $::allowtags {
-            if {[lsearch $::tags $tag] >= 0} {
-                incr matched
-            }
-        }
-        if {$matched < 1} {
-            incr ::num_aborted
-            send_data_packet $::test_server_fd ignore $name
-            return
-        }
+    set tags [concat $::tags $tags]
+    if {![tags_acceptable $tags err]} {
+        incr ::num_aborted
+        send_data_packet $::test_server_fd ignore "$name: $err"
+        return
     }
 
     incr ::num_tests
     set details {}
     lappend details "$name in $::curfile"
 
-    # set a cur_test global to be logged into new servers that are spown
+    # set a cur_test global to be logged into new servers that are spawn
     # and log the test name in all existing servers
     set prev_test $::cur_test
     set ::cur_test "$name in $::curfile"
-    if {!$::external} {
+    if {$::external} {
+        catch {
+            set r [redis [srv 0 host] [srv 0 port] 0 $::tls]
+            catch {
+                $r debug log "### Starting test $::cur_test"
+            }
+            $r close
+        }
+    } else {
         foreach srv $::servers {
             set stdout [dict get $srv stdout]
             set fd [open $stdout "a+"]
@@ -162,6 +187,7 @@ proc test {name code {okpattern undefined} {options undefined}} {
 
     send_data_packet $::test_server_fd testing $name
 
+    set test_start_time [clock milliseconds]
     if {[catch {set retval [uplevel 1 $code]} error]} {
         set assertion [string match "assertion:*" $error]
         if {$assertion || $::durable} {
@@ -189,7 +215,8 @@ proc test {name code {okpattern undefined} {options undefined}} {
     } else {
         if {$okpattern eq "undefined" || $okpattern eq $retval || [string match $okpattern $retval]} {
             incr ::num_passed
-            send_data_packet $::test_server_fd ok $name
+            set elapsed [expr {[clock milliseconds]-$test_start_time}]
+            send_data_packet $::test_server_fd ok $name $elapsed
         } else {
             set msg "Expected '$okpattern' to equal or match '$retval'"
             lappend details $msg
